<?php
declare(strict_types=1);

namespace Everyware\Plugin\GoogleAnalytics;

use Everyware\Plugin\GoogleAnalytics\Models\Credentials;
use Google_Client;
use Google_Service_Analytics;
use Infomaker\Everyware\Base\OpenContent\OpenContentProvider;
use Infomaker\Everyware\Base\OpenContent\QueryBuilder;
use Infomaker\Everyware\Support\NewRelicLog;
use Infomaker\Everyware\Support\Storage\Cache;
use Infomaker\Everyware\Support\Str;

class GoogleAnalyticsClient
{

    /**
     * @var string
     */
    private const NOT_VALID_CREDENTIALS = 'not_valid_credentials';
    /**
     * @var int
     */
    private const DEFAULT_TIMEOUT = 10;

    /**
     * Cache interval in minutes.
     * @var int
     */
    private const GA_CACHE_INTERVAL = 10;

    /**
     * @var string
     */
    private const GA_CACHE_KEY = 'ew_fetch_google_analytics';
    /**
     * @var string
     */
    private const GA_VALID_CREDENTIALS = 'ew_valid_ga_credentials';

    /**
     * The maximum number of articles from Google Analytics.
     *
     * @var integer
     */
    private static $limit = 30;

    /**
     * Google Analytics options
     *
     * @var array
     */
    protected static $options = [
        'dimensions' => 'ga:pagePath',
        'sort' => '-ga:uniquePageviews',
        'filters' => 'ga:pagePath=~\/sport\/',
        'max-results' => 20,
    ];

    /** @var SettingsTab */
    private $settings;

    /**
     * Contains custom settings
     *
     * @var array
     */
    private $customSettings = [];

    /**
     * GoogleAnalyticsClient constructor.
     *
     * @param PluginSettings $settings
     */
    public function __construct(PluginSettings $settings)
    {
        $this->settings = $settings;
    }

    /**
     * Authorizes to Google Analytics.
     *
     * @param Credentials $credentials
     *
     * @return Google_Client
     */
    private function getAuthorizedClient(Credentials $credentials): ?Google_Client
    {
        try {
            $client = new Google_Client();
            $client->setApplicationName($credentials->project_id);
            $client->setAuthConfig($credentials->toArray());
            $client->setScopes($credentials->scopes);

            return $client;
        } catch (\Exception $e) {
            return null;
        }
    }

    /**
     * Initiates the data fetching.
     *
     * @return array
     */
    public function fetchData(): array
    {
        $credentials = $this->getCredentials();

        if ( ! $this->hasViewId() || ! $this->isValidCredentials($credentials)) {
            return [];
        }

        $client = $this->getAuthorizedClient($credentials);

        if ( ! $client instanceof Google_Client) {
            $this->invalidateCredentials($credentials);

            return [];
        }

        // Get data from GA
        $gaData = $this->getAnalytics($client);

        if ($gaData === null) {
            return [];
        }

        // Extract uuids from GA data
        $uuids = $this->extractUuids($gaData);

        // Get articles and store result
        $articles = (\count($uuids) === 0)
            ? []
            : $this->getArticleByUuids($uuids);

        $this->storeData($articles);

        return $articles;
    }

    /**
     * Get articles based of uuids
     *
     * @param array $uuids
     *
     * @return array
     */
    private function getArticleByUuids(array $uuids): array
    {
        $uuids = \array_unique($uuids, \SORT_REGULAR);

        $query = QueryBuilder::where('uuid', $uuids);

        $provider = OpenContentProvider::setup([
            'q' => $query->buildQueryString(),
            'contenttypes' => ['Article'],
            'limit' => \count($uuids),
        ]);
        $articles = $provider->queryWithRequirements();

        return $this->sortArticlesAfterUuidArray($uuids, $articles);
    }

    /**
     * @param array $uuids
     * @param       $articles
     *
     * @return array
     */
    private function sortArticlesAfterUuidArray(array $uuids, $articles): array
    {
        $sortedArticles = [];
        /** @var \OcArticle $article */
        foreach ($articles as $article) {
            /** @var int|string $position */
            $position = \array_search($article->get_value('uuid'), $uuids, true);
            if ($position === false) {
                continue;
            }
            $sortedArticles[$position] = $article;
        }

        \ksort($sortedArticles);

        return $sortedArticles;
    }

    /**
     * Extract uuids from data
     *
     * @param $data
     *
     * @return array
     */
    private function extractUuids($data): array
    {
        $uuids = [];
        $oldUrls = [];

        foreach ($data as $key => $article) {
            $url = $article[0];
            $postID = \url_to_postid($url);

            // If 0 we found nothing and have to check for old url in OC
            if ($postID === 0) {
                $oldUrls[$key] = $url;
            } else {
                $uuid = \get_post_meta($postID, 'oc_uuid', true);

                if ($uuid !== false && $uuid !== '') {
                    $uuids[$key] = $uuid;
                }
            }
        }

        if ( ! empty($oldUrls)) {
            $query = QueryBuilder::where('OriginalUrl', $oldUrls);

            $provider = OpenContentProvider::setup([
                'q' => $query->buildQueryString(),
                'contenttypes' => ['Article'],
            ]);

            $provider->setPropertyMap('Article');

            $articles = $provider->queryWithRequirements();

            if (\is_array($articles)) {
                \array_map(function ($article) use ($oldUrls, &$uuids) {
                    $articleValue = \array_search($article->originalUrl, $oldUrls, true);

                    if (\is_int($articleValue)) {
                        $uuids[$articleValue] = $article->uuid;
                    }
                }, $articles);

                \ksort($uuids);
                $uuids = \array_values($uuids);
            }
        }

        return $uuids;
    }

    /**
     * Get articles from specific positions
     *
     * @param int $start
     * @param int $limit
     *
     * @return array
     */
    public function getArticlesFromPosition($start = 0, $limit = 10): array
    {
        // If we have cached data we present it otherwise trigger and get data
        $articles = $this->shouldFetchData() ? $this->fetchData() : $this->getStoredData();

        if (\count($articles) === 0) {
            $articles = $this->getBackupData();
        }

        $start = ! is_numeric($start) ? 0 : (int)$start;
        $limit = $limit === null ? 10 : (int)$limit;

        return \array_slice($articles, \max($start - 1, 0), $limit);
    }

    /**
     * Fetch the data from Google Analytics.
     *
     * @param Google_Client $client
     *
     * @return array
     */
    private function getAnalytics(Google_Client $client): ?array
    {
        $options = [
            'dimensions' => 'ga:pagePath',
            'sort' => '-ga:uniquePageviews',
            'filters' => 'ga:pagePath=~^\/(?:[0-9a-zA-Z-])*\/.*',
            'max-results' => static::$limit,
        ];

        $viewId = $this->getViewId();
        $timeSpan = $this->getTimeSpan();

        try {
            $service = new Google_Service_Analytics($client);
            $data = $service->data_ga->get(
                "ga:$viewId", $timeSpan['from'], $timeSpan['to'], 'ga:uniquePageviews', $options
            );

            return $data->getRows();
        } catch (\Exception $e) {
            return [];
        }
    }

    /**
     * Fetch Credentials added in admin
     *
     * @return Credentials
     */
    public function getCredentials(): Credentials
    {
        return $this->settings->getCredentials();
    }

    /**
     * Fetch the time span to fetch the data from from admin settings
     *
     * @return array
     */
    public function getTimeSpan(): array
    {
        return $this->settings->getTimeSpan();
    }

    /**
     * Check if profile has been set in settings
     *
     * @return bool
     */
    public function hasViewId(): bool
    {
        $hasCustomViewID = isset($this->customSettings['view_id']) && Str::notEmpty($this->customSettings['view_id']);
        return $hasCustomViewID || $this->settings->hasViewId();
    }

    /**
     * @param array $options
     *
     * @return void
     */
    public function customize(array $options = []): void
    {
        $this->customSettings = $options;
    }

    /**
     * Determine if a data is stored
     *
     * @return bool
     */
    private function shouldFetchData(): bool
    {
        return Cache::get(self::GA_CACHE_KEY . '_' . $this->getViewId()) === false;
    }

    /**
     * Store backup of data
     *
     * @param array $data
     *
     * @return bool
     */
    private function storeBackup(array $data = []): bool
    {
        return Cache::set(self::GA_CACHE_KEY . '_' . $this->getViewId() . '_extended', $data, 2 * \WEEK_IN_SECONDS);
    }

    /**
     * Store data in the cache.
     *
     * @param $data
     *
     * @return boolean
     */
    private function storeData($data): bool
    {
        if ( ! empty($data)) {
            Cache::set(self::GA_CACHE_KEY . '_' . $this->getViewId(), $data, 60 * static::GA_CACHE_INTERVAL);

            return $this->storeBackup($data);
        }

        return false;
    }

    /**
     * Get stored data
     *
     * @return mixed
     */
    private function getStoredData()
    {
        if (($cached = Cache::get(self::GA_CACHE_KEY . '_' . $this->getViewId())) !== false) {
            return $cached;
        }

        return [];
    }

    /**
     * @return array|mixed
     */
    private function getBackupData()
    {
        if (($cachedBackup = Cache::get(self::GA_CACHE_KEY. '_' . $this->getViewId() . '_extended')) !== false) {
            NewRelicLog::error('Serves Google Analytics backup data');

            return $cachedBackup;
        }

        return [];
    }

    /**
     * Get view ID
     *
     * @return mixed
     */
    private function getViewId()
    {
        return $this->customSettings['view_id'] ?? $this->settings->getViewId();
    }

    /**
     * @param Credentials $credentials
     *
     * @return void
     */
    private function invalidateCredentials(Credentials $credentials): void
    {
        if ( ! empty($credentials->project_id)) {
            Cache::set(static::GA_VALID_CREDENTIALS, static::NOT_VALID_CREDENTIALS, 60 * static::GA_CACHE_INTERVAL);
        }
    }

    /**
     * @param Credentials $credentials
     *
     * @return bool
     */
    private function isValidCredentials(Credentials $credentials): bool
    {
        if (empty($credentials->project_id)) {
            return false;
        }

        return Cache::get(static::GA_VALID_CREDENTIALS, '') !== static::NOT_VALID_CREDENTIALS;
    }
}
